/* svg_text.c: Data structures for SVG text elements
 
   Copyright  2002 USC/Information Sciences Institute
  
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
  
   You should have received a copy of the GNU Library General Public
   License along with this program; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
  
   Author: Carl Worth <cworth@isi.edu>
*/

#include <string.h>

#include "svgint.h"

svg_status_t
_svg_tspan_init (svg_tspan_t *tspan)
{
    tspan->chars = NULL;
    tspan->len = 0;
    tspan->from_text = 0;
    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_tspan_init_copy (svg_tspan_t *tspan,
		     svg_tspan_t *other)
{
    tspan->len = other->len;
    if (tspan->len) {
	tspan->chars = malloc (tspan->len + 1);
	if (tspan->chars == NULL)
	    return SVG_STATUS_NO_MEMORY;
	memcpy (tspan->chars, other->chars, tspan->len);
	tspan->chars[tspan->len] = '\0';
    } else {
	tspan->chars = NULL;
    }

    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_tspan_deinit (svg_tspan_t *tspan)
{
    free (tspan->chars);
    tspan->len = 0;

    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_tspan_append_chars (svg_tspan_t	*tspan,
			const char	*chars,
			int		len)
{
    char *new_chars;

    tspan->len += len;

    new_chars = realloc (tspan->chars, tspan->len + 1);
    if (new_chars == NULL) {
	tspan->len -= len;
	return SVG_STATUS_NO_MEMORY;
    }

    if (tspan->chars == NULL)
	new_chars[0] = '\0';
    tspan->chars = new_chars;
    strncat (tspan->chars, chars, len);

    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_tspan_render (svg_tspan_t		*tspan,
		   double		x,
		   double		y,
		   svg_render_engine_t	*engine,
		   void			*closure)
{
    return (engine->render_text) (closure, x, y, tspan->chars);
}

svg_status_t
_svg_tspan_apply_attributes (svg_tspan_t	*tspan,
			     const char		**attributes)
{
#if 0
    _svg_attribute_get_lengths (attributes, "x", &tspan->x, "");
    _svg_attribute_get_lengths (attributes, "y", &tspan->y, "");
    _svg_attribute_get_lengths (attributes, "dx", &tspan->dx, "");
    _svg_attribute_get_lengths (attributes, "dy", &tspan->dy, "");
#endif

    /* XXX: What else goes here? */

    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_tspan_extents (svg_tspan_t		*tspan,
		    double		*width,
		    double		*height,
		    svg_render_engine_t	*engine,
		    void		*closure)
{
    return (engine->text_extents) (closure,
				   tspan->chars,
				   width, height);
}


svg_status_t
_svg_text_init (svg_text_t *text)
{
    _svg_group_init (&text->group);
    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_text_init_copy (svg_text_t *text,
		     svg_text_t *other)
{
    svg_status_t status;
    
    status = _svg_group_init_copy (&text->group, &other->group);
    if (status)
	return status;

    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_text_deinit (svg_text_t *text)
{
    _svg_group_deinit (&text->group);

    return SVG_STATUS_SUCCESS;
}

static svg_status_t
_svg_text_prepare_tspan (svg_element_t *element,
			 svg_render_engine_t *engine,
			 void *closure)
{
    svg_transform_t transform = element->transform;
    svg_status_t status;
    
    status = _svg_transform_render (&transform, engine, closure);
    if (status)
	return status;

    status = _svg_style_render (&element->style, engine, closure);
    if (status)
	return status;
    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_text_render (svg_text_t		*text,
		  svg_render_engine_t	*engine,
		  void			*closure)
{
    double		x, y;
    int			e;
    double		w, h;
    svg_status_t	status = SVG_STATUS_SUCCESS;
    svg_text_anchor_t	text_anchor;

    (engine->measure_position) (closure,
				&text->group.x, &text->group.y, &x, &y);

    (engine->get_text_anchor) (closure, &text_anchor);

    if (text_anchor != SVG_TEXT_ANCHOR_START) {
	_svg_text_extents (text, &w, &h, engine, closure);

	switch (text_anchor) {
	case SVG_TEXT_ANCHOR_END:
	    x -= w;
	    y -= h;
	    break;
	case SVG_TEXT_ANCHOR_MIDDLE:
	    x -= w/2;
	    y -= h/2;
	    break;
	case SVG_TEXT_ANCHOR_START:
	    break;
	}
    }

    for (e = 0; e < text->group.num_elements; e++)
    {
	status = (engine->begin_element) (closure);
	if (status)
	    return status;
	status = _svg_text_prepare_tspan (text->group.element[e],
					  engine, closure);
	if (status)
	    break;
	status = _svg_tspan_render (&text->group.element[e]->e.tspan,
				    x, y, engine, closure);
	if (status)
	    break;
	status = _svg_tspan_extents (&text->group.element[e]->e.tspan,
				     &w, &h, engine, closure);
	if (status)
	    return status;
	status = (engine->end_element) (closure);
	if (status)
	    break;
	x += w;
	y += h;
    }
    return status;
}

svg_status_t
_svg_text_apply_attributes (svg_text_t		*text,
			    const char		**attributes)
{
    _svg_attribute_get_length (attributes, "x", &text->group.x, "0");
    _svg_attribute_get_length (attributes, "y", &text->group.y, "0");

    /* XXX: What else goes here? */

    return SVG_STATUS_SUCCESS;
}

svg_status_t
_svg_text_extents (svg_text_t		*text,
		   double		*width,
		   double		*height,
		   svg_render_engine_t	*engine,
		   void			*closure)
{
    int		    e;
    double	    x = 0, y = 0;
    double	    w, h;
    svg_status_t    status;

    
    for (e = 0; e < text->group.num_elements; e++)
    {
	status = (engine->begin_element) (closure);
	if (status)
	    return status;
	status = _svg_text_prepare_tspan (text->group.element[e],
					  engine, closure);
	if (status)
	    return status;
	status = _svg_tspan_extents (&text->group.element[e]->e.tspan,
				     &w, &h, engine, closure);
	if (status)
	    return status;
	status = (engine->end_element) (closure);
	if (status)
	    return status;
	x += w;
	y += h;
    }
    *width = x;
    *height = y;
    return SVG_STATUS_SUCCESS;
}
